import pickle
class Node:

    def __init__(self, id, weight=float('inf'), name="", degree=0, category=""):
        self.id = id
        self.weight = weight
        self.name = name
        self.degree = degree
        self.category = category

    def __eq__(self, other):
        if isinstance(other, Node):
            return self.id == other.id
        return False

    def __str__(self):
        return self.id

    def __hash__(self):
        return hash(self.id)

    def serialize(self):
        return pickle.dumps(self)

    @staticmethod
    def deserialize(data):
        return pickle.loads(data)



import os
import pickle
from concurrent.futures import ThreadPoolExecutor

import numpy as np
import xgboost as xgb
from importlib.resources import files as resource_files

from pathfinder.core.feature_extractor import get_neighbors_info, get_category, get_np_array_features
from pathfinder.core.model.Node import Node
from pathfinder.core.repo.Repository import Repository


def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))


class MLRepo(Repository):

    def __init__(self, repo, degree_repo, ngd_repo):
        self.repo = repo
        self.degree_repo = degree_repo
        self.ngd_repo = ngd_repo
        self.bst_loaded = None
        self.ancestors_by_id = None
        self.category_to_idx = None
        self.edge_category_to_idx = None
        self.sorted_category_list = None
        self.node_degree_category_to_idx = None

    def load_data(self):
        pkg_files = resource_files('pathfinder.resources')
        with pkg_files.joinpath('node_degree_category_by_indices.pkl').open('rb') as f:
            self.node_degree_category_to_idx = pickle.load(f)

        with pkg_files.joinpath('sorted_category_list.pkl').open('rb') as f:
            self.sorted_category_list = pickle.load(f)

        with pkg_files.joinpath('edge_category_to_idx.pkl').open('rb') as f:
            self.edge_category_to_idx = pickle.load(f)
        self.category_to_idx = {cat_name: idx for idx, cat_name in enumerate(self.sorted_category_list)}

        with pkg_files.joinpath('ancestors_by_indices.pkl').open('rb') as f:
            self.ancestors_by_id = pickle.load(f)

        self.bst_loaded = xgb.Booster()
        self.bst_loaded.load_model(str(pkg_files.joinpath('pathfinder_xgboost_model_kg_2_10_2')))

    def get_neighbors(self, node, limit=-1):
        if limit <= 0:
            raise Exception(f"The limit:{limit} could not be negative or zero.")

        with ThreadPoolExecutor(max_workers=2) as executor:
            future_get_neighbors = executor.submit(
                get_neighbors_info,
                node.id,
                self.ngd_repo,
                self.repo,
                self.degree_repo
            )
            future_load_data = executor.submit(self.load_data)

            content_by_curie, curie_category = future_get_neighbors.result()

        if content_by_curie is None:
            return []
        curie_category_onehot = get_category(curie_category.split(":")[-1], self.category_to_idx)

        feature_list = []
        curie_list = []
        curie_degree = []
        curie_name = []
        curie_category = []
        for key, value in content_by_curie.items():
            curie_list.append(key)
            curie_degree.append(value.get('degree_by_category', {}).get('biolink:NamedThing', 0))
            curie_name.append(value.get('name', ''))
            curie_category.append(value.get('category', ''))
            feature_list.append(
                get_np_array_features(
                    value,
                    self.category_to_idx,
                    self.edge_category_to_idx,
                    curie_category_onehot,
                    self.ancestors_by_id,
                    self.node_degree_category_to_idx
                )
            )

        feature_np = np.empty((len(feature_list), len(feature_list[0])), dtype=float)

        for i in range(len(feature_list)):
            feature_np[i] = feature_list[i]

        dtest = xgb.DMatrix(feature_np)

        scores = self.bst_loaded.predict(dtest)

        probabilities = sigmoid(scores)

        ranked_items = sorted(
            zip(curie_list, probabilities, curie_degree, curie_category, curie_name),
            key=lambda x: x[1],
            reverse=True
        )

        return [Node(
            id=item[0],
            weight=float(item[1]),
            degree=item[2],
            category=item[3],
            name=item[4]
        ) for item in ranked_items[0:limit]]

    def get_node_degree(self, node_id):
        return self.degree_repo.get_node_degree(node_id)



import multiprocessing

import queue

from pathfinder.core.constants import NODE_DEGREE_LIMIT, NEIGHBOR_LIMIT
from pathfinder.core.model.Node import Node
from pathfinder.core.model.Path import Path
from pathfinder.core.model.PathFinderModel import PathFinderModel
from pathfinder.core.repo.repo_factory import get_repo


def process_path(path_string):
    try:
        path_finder_model = PathFinderModel.deserialize(path_string)
        result = []
        if path_finder_model.path.path_limit > 0:
            last_link = path_finder_model.path.last()
            repo = get_repo(path_finder_model.repo_name, path_finder_model.plover_url, path_finder_model.ngd_url,
                            path_finder_model.degree_url)
            node_degree = repo.get_node_degree(last_link.id)
            if node_degree > NODE_DEGREE_LIMIT:
                return path_string, result, None
            neighbors = repo.get_neighbors(last_link, NEIGHBOR_LIMIT)
            for neighbor in neighbors:
                if neighbor not in path_finder_model.path.links:
                    new_path = path_finder_model.path.make_new_path(neighbor)
                    result.append(new_path.serialize())

        return path_string, result, None
    except Exception as e:
        return path_string, None, e


class BreadthFirstSearch:

    def __init__(self, repository_name, plover_url, ngd_url, degree_url, path_container, logger):
        self.repo_name = repository_name
        self.plover_url = plover_url
        self.ngd_url = ngd_url
        self.degree_url = degree_url
        self.path_container = path_container
        self.logger = logger

    def traverse(self, source_id, hops_numbers=1):
        self.logger.info(f"Expanding {source_id}")
        path_queue = queue.Queue()
        new_path = Path(hops_numbers, [Node(source_id, weight=1)])
        path_queue.put(new_path)
        self.path_container.add_new_path(new_path)

        if hops_numbers == 0:
            return

        num_cores = min(multiprocessing.cpu_count(), 4)

        with multiprocessing.Pool(num_cores) as pool:
            while not path_queue.empty():
                paths = []
                for _ in range(4 * num_cores):
                    if not path_queue.empty():
                        paths.append(PathFinderModel(self.repo_name, path_queue.get(), self.plover_url, self.ngd_url,
                                                     self.degree_url).serialize())

                new_paths_list = pool.map(process_path, paths)

                for path_string, new_paths, exception in new_paths_list:
                    if exception:
                        path_finder_model = PathFinderModel.deserialize(path_string)
                        self.logger.error(f"Path {path_finder_model.path} raised an exception: {exception}")
                        raise exception
                    else:
                        for new_path in new_paths:
                            p = Path.deserialize(new_path)
                            path_queue.put(p)
                            self.path_container.add_new_path(p)



import logging
import requests

from pathfinder.core.model.Node import Node
from pathfinder.core.repo.Repository import Repository


class PloverDBRepo(Repository):

    def __init__(self, plover_url, degree_repo):
        self.plover_url = plover_url
        self.degree_repo = degree_repo

    def get_neighbors(self, node, limit=-1):
        endpoint = "/get_neighbors"
        data = {"node_ids": [node.id], "categories": ["biolink:NamedThing"]}
        result = []
        try:
            response = requests.post(self.plover_url + endpoint, headers={'accept': 'application/json'}, json=data)
            response.raise_for_status()
            json = response.json()
            result = [Node(i) for i in json[node.id]]
        except requests.exceptions.RequestException as e:
            # log here print(f"Request error: {e}")
            pass
        except ValueError as e:
            # log here print(f"JSON decode error: {e}")
            pass
        except KeyError as e:
            # log here print(f"Key error: {e}")
            pass
        except Exception as e:
            pass
        return result

    def get_neighbors_with_edges(self, node_id_input, limit=-1):
        endpoint = "/query"
        data = {
            "edges": {
                "e": {
                    "subject": "n1",
                    "object": "n2"
                }
            },
            "nodes": {
                "n1": {
                    "ids": [
                        node_id_input
                    ]
                },
                "n2": {
                    "categories": [
                        "biolink:NamedThing"
                    ]
                }
            },
            "include_metadata": True,
            "respect_predicate_symmetry": True
        }
        try:
            response = requests.post(self.plover_url + endpoint,
                                     headers={'accept': 'application/json'}, json=data)
            response.raise_for_status()
            json = response.json()

            if len(json['nodes']['n1']) == 0 or len(json['nodes']['n2']) == 0:
                return None, None, None

            nodes = {}
            edges = {}
            for _, info in json['edges']['e'].items():
                if info[0] == node_id_input:
                    neighbor_id = info[1]
                elif info[1] == node_id_input:
                    neighbor_id = info[0]
                else:
                    continue
                nodes[neighbor_id] = {}
                nodes[neighbor_id]['name'] = json['nodes']['n2'][neighbor_id][0]
                nodes[neighbor_id]['category'] = json['nodes']['n2'][neighbor_id][1]
                edge_info = (info[2], info[3])
                if neighbor_id not in edges:
                    edges[neighbor_id] = [edge_info]
                else:
                    edges[neighbor_id].append(edge_info)
            if node_id_input in json['nodes']['n1']:
                return json['nodes']['n1'][node_id_input][1], nodes, edges
            return None, None, None
        except requests.exceptions.RequestException as e:
            logging.error("A requests error occurred: %s", e, exc_info=True)
            raise e
        except Exception as e:
            logging.error("An unexpected error occurred: %s", e, exc_info=True)
            raise e

    def get_node_degree(self, node_id):
        return self.degree_repo.get_node_degree(node_id)



import math

NGD_normalizer = 3.5e+7 * 20
log_NGD_normalizer = math.log(NGD_normalizer)


def calculate_ngd(first_element_set, second_element_set):
    if first_element_set is None or second_element_set is None:
        return None

    log_length_of_first_element = math.log(len(first_element_set))
    log_length_of_second_element = math.log(len(second_element_set))
    joint_count = len(first_element_set.intersection(second_element_set))

    if log_length_of_first_element == 0 or log_length_of_second_element == 0:
        return None
    elif joint_count == 0:
        return None
    else:
        try:
            return (max([log_length_of_first_element, log_length_of_second_element]) - math.log(joint_count)) / \
                (log_NGD_normalizer - min([log_length_of_first_element, log_length_of_second_element]))
        except ValueError:
            return None


def calculate_ngd(log_of_length_of_first_pmid, length_of_second_pmid, length_of_intersection):
    if log_of_length_of_first_pmid == 0 or length_of_second_pmid == 0 or length_of_intersection == 0:
        return None

    log_of_length_of_second_element = math.log(length_of_second_pmid)
    try:
        return (max([log_of_length_of_first_pmid, log_of_length_of_second_element]) - math.log(
            length_of_intersection)) / \
            (log_NGD_normalizer - min([log_of_length_of_first_pmid, log_of_length_of_second_element]))
    except ValueError:
        return None



class PathListToGraphConverter:

    def __init__(self, source_node_name, destination_node_name):
        self.source_node_name = source_node_name
        self.destination_node_name = destination_node_name
        self.counter_to_generate_new_node_name = 0
        self.counter_to_generate_new_edge_name = 0
        self.node_name = "n00_"
        self.edge_name = "e00_"

    def generate_new_node_name(self):
        self.counter_to_generate_new_node_name += 1
        return f"{self.node_name}{self.counter_to_generate_new_node_name}"

    def generate_new_edge_name(self):
        self.counter_to_generate_new_edge_name += 1
        return f"{self.edge_name}{self.counter_to_generate_new_edge_name}"

    def convert(self, paths):
        nodes = {}
        edges = {}
        for path in paths:
            if len(path.links) < 2:
                # todo error log
                continue
            for i in range(0, len(path.links)):
                if i == 0:
                    if path.links[i].id not in nodes:
                        nodes[path.links[i].id] = self.source_node_name
                elif i == len(path.links) - 1:
                    if path.links[i].id not in nodes:
                        nodes[path.links[i].id] = self.destination_node_name
                    edge_exist = False
                    for key, edge in edges.items():
                        if (edge[0] == self.destination_node_name and edge[1] == nodes[path.links[i - 1].id]) \
                                or (edge[1] == self.destination_node_name and edge[0] == nodes[path.links[i - 1].id]):
                            edge_exist = True
                    if not edge_exist:
                        edges[self.generate_new_edge_name()] = (nodes[path.links[i - 1].id], self.destination_node_name)
                else:
                    if path.links[i].id not in nodes:
                        nodes[path.links[i].id] = self.generate_new_node_name()
                    edge_exist = False
                    for key, edge in edges.items():
                        if (edge[0] == nodes[path.links[i].id] and edge[1] == nodes[path.links[i - 1].id]) \
                                or (edge[1] == nodes[path.links[i].id] and edge[0] == nodes[path.links[i - 1].id]):
                            edge_exist = True
                    if not edge_exist:
                        edges[self.generate_new_edge_name()] = (nodes[path.links[i - 1].id], nodes[path.links[i].id])
        nodes = {value: key for key, value in nodes.items()}
        return nodes, edges



from pathfinder.core.BidirectionalPathFinder import BidirectionalPathFinder

from pathfinder.converter.EdgeExtractorFromPloverDB import EdgeExtractorFromPloverDB
from pathfinder.converter.ResultPerPathConverter import ResultPerPathConverter


class Pathfinder:

    def __init__(
            self,
            repository_name,
            plover_url,
            ngd_url,
            degree_url,
            blocked_curies,
            blocked_synonyms,
            logger
    ):
        self.repo_name = repository_name
        self.plover_url = plover_url
        self.ngd_url = ngd_url
        self.degree_url = degree_url
        self.blocked_curies = blocked_curies
        self.blocked_synonyms = blocked_synonyms
        self.logger = logger

    def get_paths(
            self,
            src_node_id,
            dst_node_id,
            src_pinned_node,
            dst_pinned_node,
            hops_numbers=6,
            limit=500,
            category_constraints=[]
    ):
        self.logger.info(f"Model release date: 12/01/2025")
        path_finder = BidirectionalPathFinder(
            "MLRepo",
            self.plover_url,
            self.ngd_url,
            self.degree_url,
            self.logger
        )
        paths = path_finder.find_all_paths(
            src_node_id,
            dst_node_id,
            hops_numbers=6
        )

        paths = self.remove_block_list(paths, hops_numbers)

        if len(category_constraints) > 0:
            paths = self.filter_with_constraint(paths, category_constraints)

        paths = paths[:limit]
        self.logger.info(f"PathFinder found {len(paths)} paths")

        edge_extractor = EdgeExtractorFromPloverDB(
            self.plover_url
        )
        return ResultPerPathConverter(
            paths,
            src_node_id,
            dst_node_id,
            src_pinned_node,
            dst_pinned_node,
            "aux",
            edge_extractor
        ).convert(self.logger)


    def filter_with_constraint(self, paths, category_constraints):
        result = []
        for path in paths:
            path_length = len(path.links)
            if path_length > 2:
                for i in range(1, path_length - 1):
                    node = path.links[i]
                    if node.category in category_constraints:
                        result.append(path)
                        break
        return result

    def remove_block_list(self, paths, hops_numbers):
        result = []
        for path in paths:
            append = True
            path_length = len(path.links)
            if path_length > hops_numbers + 1:
                continue
            if path_length > 2:
                for i in range(1, path_length - 1):
                    node = path.links[i]
                    if node.id in self.blocked_curies:
                        append = False
                        break
                    if node.name is not None:
                        if node.name.lower() in self.blocked_synonyms:
                            append = False
                            break
                        if node.name.lower().startswith("cyp"):
                            append = False
                            break
            if append:
                result.append(path)
        return result



import numpy as np


def get_neighbors_info(curie, ngd_repo, plover_repo, degree_repo):
    curie_ngd_list = ngd_repo.get_curie_ngd(curie)
    curie_category, neighbors, edges = plover_repo.get_neighbors_with_edges(curie)
    if neighbors:
        neighbors_id = [curie for curie, _ in neighbors.items()]
        degrees_by_node = degree_repo.get_degrees_by_node(neighbors_id)
        node_pmids_length = ngd_repo.get_curies_pmid_length(neighbors_id)
        content_by_curie = {item: {'pmids': 0, 'ngd': None, 'category': None, 'edges': {}} for item in neighbors_id}
        for curie_, degree_by_category in degrees_by_node.items():
            content_by_curie[curie_]['degree_by_category'] = degree_by_category
        for curie_, num_of_pmids in node_pmids_length:
            content_by_curie[curie_]['pmids'] = num_of_pmids
        for ngd in curie_ngd_list:
            if ngd[0] in content_by_curie:
                content_by_curie[ngd[0]]['ngd'] = ngd[1]
        for node_id, value in neighbors.items():
            content_by_curie[node_id]['name'] = value['name']
            content_by_curie[node_id]['category'] = value['category']
        if edges:
            for node, categories in edges.items():
                for category, _ in categories:
                    if category not in content_by_curie[node]['edges']:
                        content_by_curie[node]['edges'][category] = 1
                    else:
                        content_by_curie[node]['edges'][category] = content_by_curie[node]['edges'][category] + 1

        return content_by_curie, curie_category
    return None, None


def get_np_array_features(
        value,
        category_to_idx,
        edge_category_to_idx,
        curie_category_onehot,
        ancestors_by_indices,
        degree_category_to_idx
):
    ngd_val = float(value["ngd"]) if value["ngd"] is not None else np.nan
    pmid_val = float(value["pmids"])
    cat_onehot = get_category(value["category"].split(":")[-1], category_to_idx)

    edge_categories = np.zeros(len(edge_category_to_idx), dtype=float)
    for category, count in value['edges'].items():
        if category:
            if category in edge_category_to_idx:
                edge_cat_idx = edge_category_to_idx[category]
                for ancestor in ancestors_by_indices[edge_cat_idx]:
                    edge_categories[ancestor] = 1

    node_degrees_feature = np.zeros(len(degree_category_to_idx), dtype=float)
    for category, count in value['degree_by_category'].items():
        if category in degree_category_to_idx:
            category_index = degree_category_to_idx[category]
            node_degrees_feature[category_index] = count

    return np.concatenate([[ngd_val, pmid_val], cat_onehot, edge_categories, curie_category_onehot, node_degrees_feature])


def get_category(cat_str, category_to_idx):
    cat_onehot = np.zeros(58, dtype=float)

    if cat_str:
        if cat_str in category_to_idx:
            cat_idx = category_to_idx[cat_str]
            cat_onehot[cat_idx] = 1.0
    return cat_onehot



import requests


class EdgeExtractorFromPloverDB:

    def __init__(self, plover_url):
        self.plover_url = plover_url
        self.pairs_to_edge_ids = {}
        self.knowledge_graph = {'edges': {}, 'nodes': {}}

    def get_extractor_url(self):
        return self.plover_url

    def get_edges(self, pairs, logger):
        cached_pairs = []
        i = 0
        while i < len(pairs):
            edge_key_1 = f"{pairs[i][0]}--{pairs[i][1]}"
            edge_key_2 = f"{pairs[i][1]}--{pairs[i][0]}"
            if edge_key_1 in self.pairs_to_edge_ids:
                cached_pairs.append(edge_key_1)
                del pairs[i]
            elif edge_key_2 in self.pairs_to_edge_ids:
                cached_pairs.append(edge_key_2)
                del pairs[i]
            else:
                i += 1

        url = self.plover_url + "/get_edges"
        query = {"pairs": pairs}
        try:
            knowledge_graph = {'edges': {}, 'nodes': {}}
            if len(pairs) != 0:
                response = requests.post(url, headers={'accept': 'application/json'}, json=query)
                json = response.json()
                self.pairs_to_edge_ids.update(json["pairs_to_edge_ids"])
                knowledge_graph = json["knowledge_graph"]
                self.knowledge_graph['edges'].update(knowledge_graph['edges'])
                self.knowledge_graph['nodes'].update(knowledge_graph['nodes'])
            for cached_pair in cached_pairs:
                list_of_edges = self.pairs_to_edge_ids[cached_pair]
                cached_nodes = cached_pair.split("--")
                if cached_nodes[0] in self.knowledge_graph['nodes']:
                    knowledge_graph['nodes'][cached_nodes[0]] = self.knowledge_graph['nodes'][cached_nodes[0]]
                if cached_nodes[1] in self.knowledge_graph['nodes']:
                    knowledge_graph['nodes'][cached_nodes[1]] = self.knowledge_graph['nodes'][cached_nodes[1]]
                for edge in list_of_edges:
                    if str(edge) in self.knowledge_graph['edges']:
                        knowledge_graph['edges'][str(edge)] = self.knowledge_graph['edges'][str(edge)]
            return knowledge_graph
        except Exception as e:
            logger.warning(f"Cannot retrieve {query} from plover DB with url: {url}, Error: {e}")
            return None



from pathfinder.core.repo.MLRepo import MLRepo
from pathfinder.core.repo.NGDRepository import NGDRepository
from pathfinder.core.repo.NGDSortedNeighborsRepo import NGDSortedNeighborsRepo
from pathfinder.core.repo.NodeDegreeRepo import NodeDegreeRepo
from pathfinder.core.repo.PloverDBRepo import PloverDBRepo


def get_repo(repo_name, plover_url, ngd_url, degree_url):
    if repo_name == "NGDSortedNeighborsRepo":
        return NGDSortedNeighborsRepo(
            PloverDBRepo(plover_url=plover_url, degree_repo=NodeDegreeRepo(degree_url)),
            NodeDegreeRepo(degree_url),
            NGDRepository(ngd_url)
        )
    elif repo_name == "MLRepo":
        return MLRepo(
            PloverDBRepo(plover_url=plover_url, degree_repo=NodeDegreeRepo(degree_url)),
            NodeDegreeRepo(degree_url),
            NGDRepository(ngd_url)
        )
    else:
        raise ValueError(f"Unknown animal_type '{repo_name}'.")



import sys
import os

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from GraphToKnowledgeGraphConverter import GraphToKnowledgeGraphConverter


class PathConverter:

    def __init__(
            self,
            path,
            qnode_1_id,
            qnode_2_id,
            aux_name,
            edge_extractor,
            score
    ):
        self.path = path
        self.qnode_1_id = qnode_1_id
        self.qnode_2_id = qnode_2_id
        self.aux_name = aux_name
        self.edge_extractor = edge_extractor
        self.score = score

    def convert(self, response):
        knowledge_graph_src_dest = GraphToKnowledgeGraphConverter(
            self.qnode_1_id,
            self.qnode_2_id,
            self.edge_extractor).convert(response, [self.path])

        analysis = {
            "resource_id": "infores:arax",
            "path_bindings": {
                "p0": [{"id": self.aux_name}]
            },
            "score": self.score
        }

        aux_graph = {
            "edges": list(knowledge_graph_src_dest["edges"].keys()),
            "attributes": []
        }

        return analysis, aux_graph, knowledge_graph_src_dest



from pathfinder.core.model.Node import Node
from pathfinder.core.repo.Repository import Repository


class NGDSortedNeighborsRepo(Repository):

    def __init__(self, repo, degree_repo, ngd_repo):
        self.repo = repo
        self.degree_repo = degree_repo
        self.ngd_repo = ngd_repo

    def get_neighbors(self, node, limit=-1):
        if limit <= 0:
            raise Exception(f"The limit:{limit} could not be negative or zero.")
        curie_ngd_list = self.ngd_repo.get_curie_ngd(node.id)
        ngd_by_curie_dict = {}
        if curie_ngd_list is not None:
            if len(curie_ngd_list) >= limit:
                return [Node(curie, ngd_value) for curie, ngd_value in
                        curie_ngd_list[0:min(limit, len(curie_ngd_list))]]
            else:
                for curie, ngd in curie_ngd_list:
                    ngd_by_curie_dict[curie] = ngd

        id_set = set(ngd_by_curie_dict.keys())
        neighbors = self.repo.get_neighbors(node, limit=limit)
        neighbors_ids = []
        for neighbor in neighbors:
            if neighbor.id not in id_set:
                neighbors_ids.append(neighbor.id)
            else:
                id_set.remove(neighbor.id)

        for remove_id in id_set:
            del ngd_by_curie_dict[remove_id]

        number_of_curie_left_to_fill_the_limit = limit - len(curie_ngd_list)

        node_pmids_length = self.ngd_repo.get_curies_pmid_length(neighbors_ids,
                                                                 number_of_curie_left_to_fill_the_limit)
        node_pmids_length = [(key, length) for key, length in node_pmids_length if length > 0]

        curies_sorted_by_their_pmids_length = sorted(node_pmids_length, key=lambda x: (x[1]), reverse=True)

        nonzero_length_pmids_curie_with_none_ngd_value = [(curie, None) for curie, _ in
                                                          curies_sorted_by_their_pmids_length]

        sorted_neighbors_tuple = sorted(ngd_by_curie_dict.items(),
                                        key=lambda x: (x[1] is None, x[1] if x[1] is not None else float('inf')))
        sorted_neighbors_tuple.extend(nonzero_length_pmids_curie_with_none_ngd_value)

        return [Node(item[0], item[1]) for item in
                sorted_neighbors_tuple[0:min(limit, len(sorted_neighbors_tuple))]]

    def get_node_degree(self, node_id):
        return self.degree_repo.get_node_degree(node_id)



import sqlite3
import json


class NodeDegreeRepo:

    def __init__(self, db_path):
        self.db_path = db_path

    def get_node_degree(self, node_id):
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        query = "SELECT neighbor_counts FROM neighbors WHERE id = ?"
        cursor.execute(query, (node_id,))
        result = cursor.fetchone()
        conn.close()
        if result:
            degree_by_biolink_type = json.loads(result[0])
            return degree_by_biolink_type.get('biolink:NamedThing', 0)
        else:
            return 0

    def get_degrees_by_node(self, curie_ids, batch_size=10000):
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        degree_dict = {}

        for i in range(0, len(curie_ids), batch_size):
            batch_ids = curie_ids[i:i + batch_size]
            placeholders = ",".join("?" for _ in batch_ids)
            query = f"SELECT id, neighbor_counts FROM neighbors WHERE id IN ({placeholders})"
            cursor.execute(query, batch_ids)

            for node_id, neighbor_counts in cursor.fetchall():
                degree_by_biolink_type = json.loads(neighbor_counts)
                degree_dict[node_id] = degree_by_biolink_type

        conn.close()

        for curie in curie_ids:
            degree_dict.setdefault(curie, {})

        return degree_dict

    def get_degree_categories(self, batch_size=10000):
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        degree_category_set = set()
        offset = 0

        while True:
            query = f"SELECT id, neighbor_counts FROM neighbors LIMIT {batch_size} OFFSET {offset}"
            cursor.execute(query)
            rows = cursor.fetchall()

            if not rows:
                break

            for node_id, neighbor_counts in rows:
                degree_by_biolink_type = json.loads(neighbor_counts)
                degree_category_set.update(degree_by_biolink_type.keys())

            offset += batch_size

        conn.close()

        return degree_category_set



import sqlite3
import ast


class NGDRepository:

    def __init__(self, db_path):
        self.db_path = db_path

    def get_curie_ngd(self, curie):
        try:
            sqlite_connection_read = sqlite3.connect(self.db_path)
            cursor = sqlite_connection_read.cursor()
            query = "SELECT ngd FROM curie_ngd WHERE curie = ?"
            cursor.execute(query, (curie,))
            row = cursor.fetchone()
            cursor.close()
            sqlite_connection_read.close()
        except Exception as e:
            raise Exception(f"{e}, db path: {self.db_path}")
        if row:
            ngds = ast.literal_eval(row[0])
            return ngds

        return []

    def get_curies_pmid_length(self, curies, limit=-1):
        try:
            sqlite_connection_read = sqlite3.connect(self.db_path)
            cursor = sqlite_connection_read.cursor()
            if limit != -1:
                query = f"""
                    SELECT curie, pmid_length
                    FROM curie_ngd
                    WHERE curie IN ({','.join('?' for _ in curies)})
                    ORDER BY pmid_length DESC
                    LIMIT {limit};
                    """
            else:
                query = f"""
                    SELECT curie, pmid_length
                    FROM curie_ngd
                    WHERE curie IN ({','.join('?' for _ in curies)})
                    ORDER BY pmid_length DESC;
                    """
            cursor.execute(query, curies)
            rows = cursor.fetchall()
            cursor.close()
            sqlite_connection_read.close()
        except Exception as e:
            raise Exception(f"{e}, db path: {self.db_path}")
        return rows



import os
import sys
import json

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from PathConverter import PathConverter


class ResultPerPathConverter:

    def __init__(
            self,
            paths,
            node_1_id,
            node_2_id,
            qnode_1_id,
            qnode_2_id,
            aux_name,
            edge_extractor,
    ):
        self.paths = paths
        self.node_1_id = node_1_id
        self.node_2_id = node_2_id
        self.qnode_1_id = qnode_1_id
        self.qnode_2_id = qnode_2_id
        self.aux_name = aux_name
        self.edge_extractor = edge_extractor

    def convert(self, logger):
        self.extract_edges(logger)

        aux_graphs = {}
        analyses = []
        knowledge_graph = {'edges': {}, 'nodes': {}}

        i = 0
        for path in self.paths:
            i = i + 1
            analysis, aux_graph , kg = PathConverter(
                path,
                self.qnode_1_id,
                self.qnode_2_id,
                f"{self.aux_name}_{i}",
                self.edge_extractor,
                path.compute_weight(),
            ).convert(logger)
            aux_graphs[f"{self.aux_name}_{i}"] = aux_graph
            analyses.append(analysis)
            knowledge_graph['edges'].update(kg['edges'])
            knowledge_graph['nodes'].update(kg['nodes'])


        result = {
            "id": "result",
            "analyses": analyses,
            "node_bindings": {
                self.qnode_1_id: [
                    {
                        "id": self.node_1_id,
                        "attributes": []
                    }
                ],
                self.qnode_2_id: [
                    {
                        "id": self.node_2_id,
                        "attributes": []
                    }
                ]
            },
            "essence": "result",
            "resource_id": "infores:arax",
        }

        return result, aux_graphs, knowledge_graph

    def extract_edges(self, logger):
        pairs = set()
        for path in self.paths:
            n1 = path.links[0]
            for i in range(1, len(path.links)):
                n2 = path.links[i]
                if f"{n2.id}--{n1.id}" not in pairs:
                    pairs.add(f"{n1.id}--{n2.id}")
                n1 = n2
            if len(pairs) > 200:
                pair_list = []
                for pair in pairs:
                    s = pair.split("--")
                    pair_list.append([s[0], s[1]])
                self.edge_extractor.get_edges(pair_list, logger)
                pairs = set()
        if len(pairs) > 0:
            pair_list = []
            for pair in pairs:
                s = pair.split("--")
                pair_list.append([s[0], s[1]])
            self.edge_extractor.get_edges(pair_list, logger)



import pickle


class PathFinderModel:

    def __init__(self, repo_name, path, plover_url, ngd_url, degree_url):
        self.repo_name = repo_name
        self.path = path
        self.plover_url = plover_url
        self.ngd_url = ngd_url
        self.degree_url = degree_url

    def serialize(self):
        return pickle.dumps(self)

    @staticmethod
    def deserialize(data):
        return pickle.loads(data)



class PathContainer:

    def __init__(self):
        self.path_dict = {}

    def add_new_path(self, new_path):
        if new_path.links:
            end_link = new_path.links[-1]
            if end_link not in self.path_dict:
                self.path_dict[end_link] = []
            self.path_dict[end_link].append(new_path)



import sys
import os

sys.path.append(os.path.dirname(os.path.abspath(__file__)) + "/../")
from converter.PathListToGraphConverter import PathListToGraphConverter

class GraphToKnowledgeGraphConverter:

    def __init__(
            self,
            qnode_1_id,
            qnode_2_id,
            edge_extractor
    ):
        self.qnode_1_id = qnode_1_id
        self.qnode_2_id = qnode_2_id
        self.edge_extractor = edge_extractor

    def convert(self, response, paths):
        nodes, edges = PathListToGraphConverter(self.qnode_1_id, self.qnode_2_id).convert(paths)

        pairs = []
        for _, edge in edges.items():
            pairs.append([nodes[edge[0]], nodes[edge[1]]])

        knowledge_graph = self.edge_extractor.get_edges(pairs, response)

        return knowledge_graph



from abc import ABC, abstractmethod


class Repository(ABC):

    @abstractmethod
    def get_neighbors(self, node, limit):
        pass

    @abstractmethod
    def get_node_degree(self, node_id):
        pass


import math
from concurrent.futures import ThreadPoolExecutor

from pathfinder.core.BreadthFirstSearch import BreadthFirstSearch
from pathfinder.core.model.Node import Node
from pathfinder.core.model.Path import Path
from pathfinder.core.model.PathContainer import PathContainer


class BidirectionalPathFinder:

    def __init__(self, repository_name, plover_url, ngd_url, degree_url, logger):
        self.repo_name = repository_name
        self.plover_url = plover_url
        self.ngd_url = ngd_url
        self.degree_url = degree_url
        self.logger = logger

    def find_all_paths(self, node_id_1, node_id_2, hops_numbers=1):
        self.logger.info("Finding paths process has started")
        result = set()
        if hops_numbers == 0:
            return result
        if node_id_1 == node_id_2:
            return result

        hops_numbers_1 = math.floor((hops_numbers + 1) / 2)
        hops_numbers_2 = math.floor(hops_numbers / 2)

        path_container_1 = PathContainer()
        bfs_1 = BreadthFirstSearch(self.repo_name, self.plover_url, self.ngd_url, self.degree_url, path_container_1,
                                   self.logger)

        path_container_2 = PathContainer()
        bfs_2 = BreadthFirstSearch(self.repo_name, self.plover_url, self.ngd_url, self.degree_url, path_container_2,
                                   self.logger)

        with ThreadPoolExecutor(max_workers=2) as ex:
            f1 = ex.submit(bfs_1.traverse, node_id_1, hops_numbers_1)
            f2 = ex.submit(bfs_2.traverse, node_id_2, hops_numbers_2)

            f1.result()
            f2.result()

        intersection_list = path_container_1.path_dict.keys() & path_container_2.path_dict.keys()

        for node in intersection_list:
            for path_1 in path_container_1.path_dict[node]:
                for path_2 in path_container_2.path_dict[node]:
                    temp_path_1 = [Node(link.id, link.weight, link.name, link.degree, link.category) for link in
                                   path_1.links]
                    temp_path_2 = []
                    temp_path_1[-1].weight = (temp_path_1[-1].weight + path_2.links[-1].weight) / 2
                    for i in range(len(path_2.links) - 2, -1, -1):
                        n2 = Node(path_2.links[i].id, path_2.links[i].weight, path_2.links[i].name,
                                  path_2.links[i].degree, path_2.links[i].category)
                        temp_path_2.append(n2)
                    temp_path_1.extend(temp_path_2)
                    if len(temp_path_1) == len(set(temp_path_1)):
                        result.add(Path(0, temp_path_1))

        result = sorted(list(result), key=lambda path: path.compute_weight(), reverse=True)

        return result



import math
import pickle

from pathfinder.core.model.Node import Node


class Path:

    def __init__(self, path_limit, links=None):
        if links is None:
            links = list()
        self.path_limit = path_limit
        self.links = links

    def __eq__(self, other):
        if isinstance(other, Path):
            return str(self) == str(other)
        return False

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        result = ""
        for link in self.links:
            result += str(link)
        return result

    def compute_weight(self):
        weight = 0
        degree_sum = 0
        for link in self.links:
            if link.weight == float('inf') or link.weight is None:
                return float('inf')
            degree_sum += link.degree
            weight += link.weight

        if degree_sum > 1:
            weight /= math.log(degree_sum, 10)

        return weight / len(self.links)

    def make_new_path(self, last_link):
        new_links = [Node(link.id, link.weight, link.name, link.degree, link.category) for link in self.links]
        new_links.append(last_link)
        return Path(self.path_limit - 1, new_links)

    def last(self):
        if len(self.links) == 0:
            raise Exception("Path is empty.")
        return self.links[-1]

    def path_src_to_id(self, id):
        new_links = []
        for link in self.links:
            new_links.append(Node(link.id, link.weight, link.name, link.degree, link.category))
            if link.id == id:
                break
        return Path(len(new_links) - 1, new_links)

    def path_id_to_dest(self, id):
        new_links = []
        for link in reversed(self.links):
            new_links.append(Node(link.id, link.weight, link.name, link.degree, link.category))
            if link.id == id:
                break
        new_links.reverse()
        return Path(len(new_links) - 1, new_links)

    def serialize(self):
        return pickle.dumps(self)

    @staticmethod
    def deserialize(data):
        return pickle.loads(data)
